﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace MiniStm32JoeC_CoreLibrary.Components {
    public partial class SubUC_PowerSupply : UserControl, iSubUcItems {
        MiniStm32BoardJoeC _board;
        PinInfo _pininfoOut;
        PinInfo _pininfoIn;
        public SubUC_PowerSupply(MiniStm32BoardJoeC referenceBoard, string[] para) {
            InitializeComponent();
            _board = referenceBoard;
            //0=cmd,1=loc,2=label,3=pin,4=secondPara
            txt_Label.Text = para[2];
            label_pinOut.Text = para[3];
            _pininfoOut = new PinInfo(para[3]);
            _pininfoOut.EnumPinMode = PinMode.OP_Out_PushPull;
            cb_ValueType.SelectedIndex = 0;
            string[] para2 = para[4].Split('|');
            label_pinIn.Text = para2[0];
            _pininfoIn = new PinInfo(para2[0]);
            _pininfoIn.EnumPinMode = PinMode.IA_In_Analog;
        }

        public string GetSetting() {
            int pinstate = 0;
            if (btn_Set_Low.BackColor != Color.Gainsboro) { pinstate = 1; }
            if (btn_Set_High.BackColor != Color.Gainsboro) { pinstate = 2; }
            //0=cmd,1=loc,2=label,3=pinOut,
            //4=secondPara (0=pinIn,1=SenseType,2=OutState)
            return $"POWER#{Location.X}:{Location.Y}#{txt_Label.Text}#{_pininfoOut.PinName}#{_pininfoIn.PinName}|{cb_ValueType.SelectedIndex}|{pinstate}#";
        }
        public void SetSetting(string[] para) {
            try {
                string[] para2 = para[4].Split('|');
                _pininfoIn = new PinInfo(para2[0]);
                _pininfoIn.EnumPinMode = PinMode.IA_In_Analog;
                cb_ValueType.SelectedIndex = int.Parse(para2[1]);
                switch (para2[2]) {
                    case "1":
                        btn_Set_Low.BackColor = Color.PaleGreen;
                        break;
                    case "2":
                        btn_Set_High.BackColor = Color.Salmon;
                        break;
                }
            } catch (Exception) {
                txt_value.Text = "-";
            }
        }
        public void Read(bool forceRead) {
            if (!chk_Active.Checked || forceRead) {
                return;
            }
            try {
                ushort counts = _board.GetPinAdc(_pininfoIn.PinCfg);
                switch (cb_ValueType.SelectedIndex) {
                    case 1:
                        double voltCountStep = _board.ReferenceVoltage / 4095.0;
                        double volt = voltCountStep * (double)counts;
                        txt_value.Text = $"{Math.Round(volt,2)} V";
                        break;
                    case 2:
                        double voltCountStep2 = _board.ReferenceVoltage / 4095.0;
                        double volt2 = voltCountStep2 * (double)counts;
                        txt_value.Text = $"{Math.Round(volt2*1000, 0)} mA";
                        break;
                    default:
                        txt_value.Text = counts.ToString();
                        break;
                }
                txt_value.BackColor = Color.White;
            } catch (Exception) {
                txt_value.BackColor = Color.Fuchsia;
            }
        }
        public void WriteConfigToBoard() {
            try {
                if (btn_Set_Low.BackColor != Color.Gainsboro) {
                    btn_Set_Low_Click(null, null);
                }
                if (btn_Set_High.BackColor != Color.Gainsboro) {
                    btn_Set_High_Click(null, null);
                }
                _board.SetPinLine(new string[] { _pininfoIn.PinCfg }, PinSetState.PinState_In_Analog);
                Read(true);
            } catch (Exception) {
                txt_value.BackColor = Color.Fuchsia;
            }
        }

        public void EditMode(bool isEditable) {
            txt_Label.Enabled = isEditable;
        }
        void ResetPins() {
            btn_Set_Low.BackColor = Color.Gainsboro;
            btn_Set_High.BackColor = Color.Gainsboro;
        }
        void btn_Set_Low_Click(object sender, EventArgs e) {
            try {
                ResetPins();
                _board.SetDigitalPin(_pininfoOut, false);
                btn_Set_Low.BackColor = Color.PaleGreen;
            } catch (Exception) {
                btn_Set_Low.BackColor = Color.Fuchsia;
            }
        }

        void btn_Set_High_Click(object sender, EventArgs e) {
            try {
                ResetPins();
                _board.SetDigitalPin(_pininfoOut, true);
                btn_Set_High.BackColor = Color.Salmon;
            } catch (Exception) {
                btn_Set_High.BackColor = Color.Fuchsia;
            }
        }
    }
}
