﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace MiniStm32JoeC_CoreLibrary.Views_Ext {
    public partial class UC_DPSxxxx : UserControl {
        MiniStm32BoardJoeC _board;
        SerialPort SP = null;
        public UC_DPSxxxx() {
            InitializeComponent();
            cb_device.SelectedIndex = 0;
        }
        public void Init(MiniStm32BoardJoeC miniStm32) {
            _board = miniStm32;
        }
        public void InitSP(SerialPort serialPort) {
            SP = serialPort;
            rad_conn_DirectSerial.Enabled = true;
        }
        //bool _closing = false;
        //public void Close() {
        //    _closing = true;
        //}
        void WaitforBusy(int timeoutInSeconds) {
            DateTime dtTimeout = DateTime.Now.AddSeconds(timeoutInSeconds);
            while (_board.IsBusy) {
                if (DateTime.Now.Ticks > dtTimeout.Ticks) {
                    throw new Exception($"WaitforBusy timeout after '{timeoutInSeconds}'");
                }
                Application.DoEvents();
                System.Threading.Thread.Sleep(10);
            }
        }


        #region Modbus
        public int SlaveAddress { get; set; } = 1;
        public ushort[] ReadHoldingRegisters(ushort address, ushort count) {
            var data = new byte[2 + 2];
            data[0] = (byte)(address >> 8);
            data[1] = (byte)address;
            data[2] = (byte)(count >> 8);
            data[3] = (byte)count;

            var rx = WriteReadFrame(0x03, data ,2 + 2 * count);
            if (rx[0] != 0x03) {
                throw new Exception($"Modbus: false function code '0x{rx[0]:X02}'");
            }
            if (rx[1] != (2 * count)) {
                throw new Exception($"Modbus: false length '0x{rx[1]:X02}'");
            }

            var registers = new ushort[rx[1] / 2];
            for (int i = 0; i < registers.Length; i++) {
                registers[i] = toUshort(rx, 2 + i * 2);
            }

            return registers;
        }

        public void WriteSingleRegister(ushort address, ushort value) {
            var data = new byte[4];
            data[0] = (byte)(address >> 8);
            data[1] = (byte)address;
            data[2] = (byte)(value >> 8);
            data[3] = (byte)value;

            var rx = WriteReadFrame(0x06, data, 5);
            if (rx[0] != 0x06) {
                throw new Exception($"Modbus: false function code '0x{rx[0]:X02}'");
            }
            int getAdd = toUshort(rx, 1);
            if (getAdd != address) {
                throw new Exception($"Modbus: false address '0x{getAdd:X02}'");
            }
            int getLen = toUshort(rx, 3);
            if (getLen != value) {
                throw new Exception($"Modbus: false value '0x{getLen:X02}'");
            }
            txt_DPS_log.AppendText("OK");
        }
        public void WriteMultipleRegisters(ushort address, ushort[] values) {
            var data = new byte[5 + 2 * values.Length];
            data[0] = (byte)(address >> 8);
            data[1] = (byte)address;
            data[2] = (byte)(values.Length >> 8);
            data[3] = (byte)values.Length;
            data[4] = (byte)(values.Length * 2);

            for (int i = 0; i < values.Length; i++) {
                data[5 + i * 2] = (byte)(values[i] >> 8);
                data[6 + i * 2] = (byte)(values[i] >> 0);
            }

            var rx = WriteReadFrame(0x10, data, 5);
            if (rx[0] != 0x10) {
                throw new Exception($"Modbus: false function code '0x{rx[0]:X02}'");
            }
            int getAdd = toUshort(rx, 1);
            if (getAdd != address) {
                throw new Exception($"Modbus: false address '0x{getAdd:X02}'");
            }
            int getLen = toUshort(rx, 3);
            if (getLen != values.Length) {
                throw new Exception($"Modbus: false length '0x{getLen:X02}'");
            }
        }
        public byte[] WriteReadFrame(byte function, byte[] tx, int bytes) {
            if (tx.Length > 252) {
                throw new Exception($"Modbus: too much bytes ({tx.Length})");
            }

            var buff = new byte[2 + tx.Length + 2];
            buff[0] = (byte)SlaveAddress;
            buff[1] = function;
            tx.CopyTo(buff, 2);

            ushort crc = CalculateCrc(buff, buff.Length - 2);
            buff[buff.Length - 2] = (byte)(crc >> 0);
            buff[buff.Length - 1] = (byte)(crc >> 8);

            //1 6 0 9 0 0 89 200
            //1 3 0 0 0 13 132 15
            byte[] buffer = null;
            if (rad_conn_board_Uart3.Checked) {
                buffer = _board.UartWriteRead(3, buff, bytes + 3, 1);
            }
            if (rad_conn_DirectSerial.Checked) {
                SP.Write(buff, 0, buff.Length);
                buffer = ReadDataDirectSerial(bytes + 3);
            }

            if (buffer.Length < 3) {
                throw new Exception($"Modbus: get not enough bytes ({buffer.Length})");
            }
            if (buffer[0] != (byte)SlaveAddress) {
                throw new Exception($"Modbus: false SlaveAdd ({buffer[0]})");
            }

            crc = CalculateCrc(buffer, buffer.Length - 2);
            if (buffer[buffer.Length - 2] != (byte)(crc >> 0) ||
                buffer[buffer.Length - 1] != (byte)(crc >> 8)) {
                throw new Exception($"Modbus: CRC error");
            }
            var response = new byte[buffer.Length - 3];
            Array.Copy(buffer, 1, response, 0, buffer.Length - 3);
            return response;

        }

        byte[] ReadDataDirectSerial(int bytes) {
            var buffer = new List<byte>(bytes);

            try {
                for (int i = 0; i < bytes; i++) {
                    var ch = SP.ReadByte();
                    if (ch < 0)
                        break;

                    buffer.Add((byte)ch);
                }
            } catch (TimeoutException) {
            }

            return buffer.ToArray();
        }
        ushort toUshort(byte[] buffer, int offset) {
            return (ushort)((buffer[offset] << 8) | buffer[offset + 1]);
        }
        ushort CalculateCrc(byte[] buffer, int length) {
            ushort crc = 0xFFFF;

            for (int i = 0; i < length; i++) {
                crc = (ushort)(crc ^ buffer[i]);
                for (int j = 0; j < 8; j++) {
                    if ((crc & 1) != 0) {
                        crc = (ushort)((crc >> 1) ^ 0xA001);
                    } else {
                        crc = (ushort)(crc >> 1);
                    }
                }
            }

            return crc;
        }
        #endregion

        void btn_DPS_OutOn_Click(object sender, EventArgs e) {
            try {
                switch (cb_device.SelectedIndex) {
                    case 1: WriteSingleRegister(18, 1); break;
                    default: WriteSingleRegister(0x09, 1); break;
                }
            } catch (Exception ex) {
                txt_DPS_log.AppendText("EX: " + ex.Message);
            }
        }

        void btn_DPS_OutOff_Click(object sender, EventArgs e) {
            try {
                switch (cb_device.SelectedIndex) {
                    case 1: WriteSingleRegister(18, 0); break;
                    default: WriteSingleRegister(0x09, 0); break;
                }
            } catch (Exception ex) {
                txt_DPS_log.AppendText("EX: " + ex.Message);
            }
        }

        private void btn_DPS_readState_Click(object sender, EventArgs e) {
            try {
                ReadDpsStatus();
            } catch (Exception ex) {
                txt_DPS_log.AppendText("EX: " + ex.Message);
            }
        }
        void ReadDpsStatus() {
            txt_DPS_log.Text = "";
            ushort[] rx = ReadHoldingRegisters(0x00, 13);
            txt_DPS_log.AppendText($"setVolt: {rx[0] * 0.01}\r\n");
            txt_DPS_log.AppendText($"setCurr: {rx[1] * 0.001}\r\n");
            txt_DPS_log.AppendText($"outVolt: {rx[2] * 0.01}\r\n");
            txt_DPS_log.AppendText($"outCurr: {rx[3] * 0.001}\r\n");
            txt_DPS_log.AppendText($"outPow: {rx[4] * 0.01}\r\n");
            txt_DPS_log.AppendText($"inputVolt: {rx[5] * 0.01}\r\n");
            txt_DPS_log.AppendText($"locked: {rx[6]}\r\n");
            txt_DPS_log.AppendText($"protection: {rx[7]}\r\n"); //0=none,1=OVP,2=CVP,3=OPP
            txt_DPS_log.AppendText($"regulation: {rx[8]}\r\n"); //0=CVolt,1=CCurr
            txt_DPS_log.AppendText($"output: {rx[9]}\r\n");
            txt_DPS_log.AppendText($"bright: {rx[10]}\r\n");
            txt_DPS_log.AppendText($"MODEL: {rx[11]}\r\n");
            txt_DPS_log.AppendText($"VERSON: {rx[12]}\r\n");
        }

        void track_DPS_Bright_Scroll(object sender, EventArgs e) {
            SetBrightness(track_DPS_Bright.Value);
        }

        void track_DPS_Bright_MouseDown(object sender, MouseEventArgs e) {
            SetBrightness(track_DPS_Bright.Value);
        }
        void SetBrightness(int intensity) {
            try {
                WriteSingleRegister(0x0A, (ushort)intensity);
            } catch (Exception ex) {
                txt_DPS_log.AppendText("EX: " + ex.Message);
            }
        }

        void num_DPS_SetVolt_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode != Keys.Enter) {
                return;
            }
            UpdateSetValues();
        }
        void num_DPS_SetCurr_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode != Keys.Enter) {
                return;
            }
            UpdateSetValues();
        }
        void UpdateSetValues() {
            try {
                ushort V = (ushort)((float)num_DPS_SetVolt.Value / 0.01f);
                ushort A = (ushort)((float)num_DPS_SetCurr.Value / 0.001f);
                WriteMultipleRegisters(0x00, new ushort[] { V, A});
            } catch (Exception ex) {
                txt_DPS_log.AppendText("EX: " + ex.Message);
            }
        }

        private void btn_dps_clearBuffer_Click(object sender, EventArgs e) {
            if (rad_conn_board_Uart3.Checked) {
                _board.UartRead(3, 50, 0);
            }
            if (rad_conn_DirectSerial.Checked) {
                SP.ReadExisting();
            }
        }

        private void btn_DPS_readView_Click(object sender, EventArgs e) {

            try {
                txt_DPS_log.Text = "";
                ushort[] rx = ReadHoldingRegisters(2, 3);
                txt_DPS_log.AppendText($"outVolt: {rx[0] * 0.01}\r\n");
                txt_DPS_log.AppendText($"outCurr: {rx[1] * 0.001}\r\n");
                txt_DPS_log.AppendText($"outPow: {rx[2] * 0.01}\r\n");
            } catch (Exception ex) {
                txt_DPS_log.AppendText("EX: " + ex.Message);
            }
        }

        void btn_regRead_Read_Click(object sender, EventArgs e) {
            try {
                //13 -> 300 (30.0 °C)
                ushort tx = ushort.Parse(txt_regRead_add.Text);
                ushort[] rx = ReadHoldingRegisters(tx, (ushort)num_regRead_items.Value);
                if (rx.Length == 1) {
                    txt_DPS_log.AppendText($"Reg R: {tx} -> {rx[0]}\r\n");
                } else {
                    for (int i = 0; i < rx.Length; i++) {
                        txt_DPS_log.AppendText($"Reg R: {tx+i} -> {rx[i]}\r\n");
                    }
                }
            } catch (Exception ex) {
                txt_DPS_log.AppendText("EX: " + ex.Message);
            }
        }

        void txt_regWrite_add_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode != Keys.Enter) {
                return;
            }
            try {
                //15->1 = locked 0 = normal
                //18->0 = out off, 1= out on
                ushort txA = ushort.Parse(txt_regWrite_add.Text);
                ushort txD = ushort.Parse(txt_regWrite_data.Text);
                WriteSingleRegister(txA, txD);
                txt_DPS_log.AppendText($"Reg W: {txA} -> {txD}\r\n");
            } catch (Exception ex) {
                txt_DPS_log.AppendText("EX: " + ex.Message);
            }
        }

        void txt_regWrite_data_KeyDown(object sender, KeyEventArgs e) {
            txt_regWrite_add_KeyDown(sender, e);
        }
    }
}
